#ifndef CREATURELIB_BASELIBRARY_HPP
#define CREATURELIB_BASELIBRARY_HPP

#include <Arbutils/Collections/Dictionary.hpp>
#include <Arbutils/Collections/List.hpp>
#include <Arbutils/Memory/Memory.hpp>
#include <Arbutils/Random.hpp>
#include <Arbutils/String/StringView.hpp>

namespace CreatureLib::Library {
    template <class T> class BaseLibrary {
    protected:
        ArbUt::Dictionary<u32, std::unique_ptr<const T>> _values;
        ArbUt::List<u32> _listValues;

    public:
        BaseLibrary(size_t initialCapacity = 32) : _values(initialCapacity), _listValues(initialCapacity) {}

        virtual ~BaseLibrary() noexcept { _values.Clear(); }

        inline virtual void Insert(const ArbUt::StringView& key, const T* non_null value) {
            EnsureNotNull(value)
            _values.GetStdMap().insert({key.GetHash(), std::unique_ptr<const T>(value)});
            _listValues.Append(key);
        }
        inline virtual void Insert(u32 hashedKey, const T* non_null value) {
            EnsureNotNull(value)
            _values.GetStdMap().insert({hashedKey, std::unique_ptr<const T>(value)});
            _listValues.Append(hashedKey);
        }

        inline void Delete(const ArbUt::StringView& key) noexcept {
            _values.erase(key.GetHash());
            auto k = _listValues.IndexOf(key);
            if (k.has_value()) {
                _listValues.Remove(k.value());
            }
        }
        inline void Delete(u32 hashedKey) noexcept {
            _values.Remove(hashedKey);
            auto k = _listValues.IndexOf(hashedKey);
            if (k.has_value()) {
                _listValues.Remove(k.value());
            }
        }

        std::optional<ArbUt::BorrowedPtr<const T>> TryGet(const ArbUt::BasicStringView& name) const noexcept {
            return TryGet(name.GetHash());
        }
        std::optional<ArbUt::BorrowedPtr<const T>> TryGet(u32 hashedKey) const noexcept {
            auto find = _values.GetStdMap().find(hashedKey);
            if (find == _values.GetStdMap().end())
                return {};
            return std::get<1>(*find);
        }

        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> Get(const ArbUt::BasicStringView& name) const {
            return _values.Get(name.GetHash());
        }
        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> Get(u32 hashedKey) const { return _values.Get(hashedKey); }

        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> operator[](const ArbUt::BasicStringView& name) const {
            return Get(name);
        }
        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> operator[](u32 hashedKey) const { return Get(hashedKey); }
        [[nodiscard]] inline const ArbUt::Dictionary<u32, const std::unique_ptr<const T>>&
        GetIterator() const noexcept {
            return _values;
        }

        using const_iterator = typename std::unordered_map<u32, ArbUt::BorrowedPtr<const T>>::const_iterator;

        inline const_iterator begin() const {
            return reinterpret_cast<const std::unordered_map<u32, ArbUt::BorrowedPtr<const T>>&>(_values.GetStdMap())
                .begin();
        }
        inline const_iterator end() const {
            return reinterpret_cast<const std::unordered_map<u32, ArbUt::BorrowedPtr<const T>>&>(_values.GetStdMap())
                .end();
        }

        [[nodiscard]] size_t GetCount() const noexcept { return _values.Count(); }
        ArbUt::BorrowedPtr<const T> GetAtIndex(size_t index) const {
            return std::next(std::begin(_values), index)->second;
        }

        inline ArbUt::BorrowedPtr<const T> GetRandomValue(ArbUt::Random& rand) const noexcept {
            auto i = rand.Get(_listValues.Count());
            return _values[_listValues[i]];
        }
    };
}

#endif // CREATURELIB_BASELIBRARY_HPP
